/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "global.h"
#include "defines.h"
//STATIC #include "rfutil.h"
#include "strings.h"


#define MAX_PARAMS 200

#define ARG_FILENAME        0
#define ARG_INTEGER         1
#define ARG_FLOAT           2
#define ARG_NONE            3
#define ARG_INTLIST         4
#define ARG_FLOATLIST       5

typedef struct
{
    struct param_desc
    {
        char *name;
        char *default_string;
        int type;
        void *value;
    } params[MAX_PARAMS];
    int num;
} param_list;


/* Adds new parameter to the list of valid parameters */
static void add_param_to_list(param_list *list, char *name, char *default_string, int type, void *value)
{
    list->params[list->num].name = name;
    list->params[list->num].default_string = default_string;
    list->params[list->num].type = type;
    list->params[list->num].value = value;
    list->num++;
}

/* Read parameter strings from a file and put them in an "argv" array */
static char **read_config_file(FILE *f, int *num_of_params)
{
    char a[2000];
    char b[10];
    int ret, ret2;
    char **argv;
    int n;

    argv = malloc((MAX_PARAMS+1)*sizeof(char*));
    if (argv == NULL)
        return NULL;

    for (n = 0; n <= MAX_PARAMS; n++)
        argv[n] = NULL;

    n = 1;
    while (n <= MAX_PARAMS)
    {    
        /* Check for double quote */
        ret2 = fscanf(f, " %[\"]", b);

        /* There is no double quote -> read next string */
        if (ret2 == 0)
        {
            ret = fscanf(f, " %s", a);

            if (ret < 1 || ret == EOF)
                break;

            if (a[0] == ';')
            {
                fscanf(f, "%[^\n]", a);
                continue;
            }
        }
        /* There is double quote -> read a string enclosed in double quotes */
        else
        {
            ret = fscanf(f, "%[^\",\n]", a);

            if (ret < 1 || ret == EOF)
                break;

            fscanf(f, "\""); /* Read trailing double quote */
        }

        argv[n] = malloc((strlen(a)+1)*sizeof(char));

        if (argv[n] == NULL)
        {
            return NULL;
        }
        strcpy(argv[n], a);
        n++;
    }

    *num_of_params = n;
    return argv;
}

/* Delete parameter strings */
static void delete_config_file_params(char **params)
{
    int n;

    for (n = 0; n <= MAX_PARAMS; n++)
    {
        if (params[n] != NULL)
            free(params[n]);
    }

    free(params);
}

/* Interpret parameter strings */
static int parse_params(int argc, char **argv, enc_params *params, param_list *list)
{
    int i, j;
    FILE *config;
    char **file_params;
    int num_file_params;

    char *tmp;
    int cnt;
    
    for (i = 1; i < argc; i++)
    {
        /* Check if parameter is in the list of known parameters */
        for (j = 0; j < list->num; j++)
        {
            if (strcmp(argv[i], list->params[j].name) == 0)
                break;
        }
        if (j == list->num)
        {
            fprintf(stderr, "Unknown parameter: %s\n", argv[i]);
            return -1;
        }

        /* Parse parameter value according to its type */
        switch (list->params[j].type)
        {

        /* Parameter value is a filename */
        case ARG_FILENAME:
            i++;
            if (i == argc)
            {
                fprintf(stderr, "No filename found for parameter: %s\n", argv[i-1]);
                return -1;
            }
            if (strcmp(argv[i-1], "-cf") == 0)
            {
                /* Read the config file and parse it recursively */
                config = fopen(argv[i], "r");
                if (config == NULL)
                {
                    fprintf(stderr, "Cannot open config file: %s\n", argv[i]);
                    return -1;
                }
                file_params = read_config_file(config, &num_file_params);
                fclose(config);
                if (file_params == NULL)
                {
                    return -1;
                }
                if (parse_params(num_file_params, file_params, params, list) < 0)
                {
                    return -1;
                }
                delete_config_file_params(file_params);
            }
            else
            {
                if (*((char **)list->params[j].value) != NULL)
                {
                    free(*((char **)list->params[j].value));
                }
                *((char **)list->params[j].value) = malloc((strlen(argv[i])+1)*sizeof(char));
                if (*((char **)list->params[j].value) == NULL)
                {
                    fprintf(stderr, "Memory allocation failed\n");
                    return -1;
                }
                strcpy(*((char **)list->params[j].value), argv[i]);
            }
            break;

        /* Parameter value is a comma-separated list of integers */
        case ARG_INTLIST:
            i++;
            if (i == argc)
            {
                fprintf(stderr, "No integer list found for parameter: %s\n", argv[i-1]);
                return -1;
            }
            cnt = 1;
            if ((tmp = strtok(argv[i], ", ")) == NULL)
            {
                fprintf(stderr,"Error reading integer list for parameter: %s\n", argv[i-1]);
                return -1;
            }
            else
            {
/*                 fprintf(stdout," Read %d\n", atoi(tmp)); */
                *(((int *)list->params[j].value)+cnt++) = atoi(tmp);
            }
            while ((tmp = strtok(NULL,", ")) != NULL)
            {
/*                 fprintf(stdout," Read %d\n", atoi(tmp)); */
                *(((int *)list->params[j].value)+cnt++) = atoi(tmp);
            }
            *(((int *)list->params[j].value)+0) = cnt-1;
            break;
        /* Parameter value is an integer */
        case ARG_INTEGER:
            i++;
            if (i == argc)
            {
                fprintf(stderr, "No value found for parameter: %s\n", argv[i-1]);
                return -1;
            }
            *((int *)list->params[j].value) = atoi(argv[i]);
            break;

        /* Parameter value is a floating point value */
        case ARG_FLOAT:
            i++;
            if (i == argc)
            {
                fprintf(stderr, "No value found for parameter: %s\n", argv[i-1]);
                return -1;
            }
            *((float *)list->params[j].value) = (float)atof(argv[i]);
            break;

        /* Parameter is a flag and does not have value */
        case ARG_NONE:
            *((int *)list->params[j].value) = 1;
            break;

        default:
            break;

        }
    }

    return 0;
}


enc_params *parse_config_params(int argc, char **argv)
{
    param_list list;
    enc_params *params;
    char *default_argv[MAX_PARAMS*2];
    int default_argc;
    int i;

    params = (enc_params *)malloc(sizeof(enc_params));
    if (params == NULL)
    {
        fprintf(stderr, "Memory allocation failed\n");
        return NULL;
    }

    memset(params, 0, sizeof(enc_params));
    memset(&list, 0, sizeof(param_list));


    add_param_to_list(&list, "-cf",        NULL, ARG_FILENAME, NULL);
    add_param_to_list(&list, "-if",        NULL, ARG_FILENAME, &params->infilestr);
    add_param_to_list(&list, "-of",        NULL, ARG_FILENAME, &params->outfilestr);
    add_param_to_list(&list, "-rf",        NULL, ARG_FILENAME, &params->reconfilestr);
    add_param_to_list(&list, "-stat",      NULL, ARG_FILENAME, &params->statfilestr);
    add_param_to_list(&list, "-n",    "1000000", ARG_INTEGER,  &params->num_frames);
    add_param_to_list(&list, "-intra_period","0", ARG_INTEGER, &params->intra_period);
    add_param_to_list(&list, "-cols",     "176", ARG_INTEGER,  &params->cols);
    add_param_to_list(&list, "-rows",     "144", ARG_INTEGER,  &params->rows);
    add_param_to_list(&list, "-qp",        "20", ARG_INTEGER,  &params->qp);
    add_param_to_list(&list, "-dqpi",       "0", ARG_INTEGER,  &params->dqpI);
    add_param_to_list(&list, "-ifreq",     "15", ARG_FLOAT,    &params->ifreq);
    add_param_to_list(&list, "-ofreq",     "15", ARG_FLOAT,    &params->ofreq);
    add_param_to_list(&list, "-trace",      "0", ARG_INTEGER,  &params->trace);
    add_param_to_list(&list, "-me_type",    "0", ARG_INTEGER,  &params->me_type);
    add_param_to_list(&list, "-me_range",  "16", ARG_INTEGER,  &params->me_range);
    add_param_to_list(&list, "-lambda",   "0.9", ARG_FLOAT,    &params->lambda_coef);
	add_param_to_list(&list, "-first_frames_skip",	"0",	 ARG_INTEGER,  &params->first_frames_skip);
    add_param_to_list(&list, "-fast_svt",   "0",ARG_INTEGER,  &params->fast_svt);
    add_param_to_list(&list, "-fast_tr4_thr1",   "0",ARG_INTEGER,  &params->fast_tr4_thr1);
    add_param_to_list(&list, "-fast_svt_thr2",   "0",ARG_INTEGER,  &params->fast_svt_thr2);
    add_param_to_list(&list, "-svt",        "0",ARG_INTEGER,  &params->svt);
    add_param_to_list(&list, "-dqpb0",      "4", ARG_INTEGER,  &params->dqpB[0]);
    add_param_to_list(&list, "-dqpb1",      "5", ARG_INTEGER,  &params->dqpB[1]);
    add_param_to_list(&list, "-dqpb2",      "6", ARG_INTEGER,  &params->dqpB[2]);
    add_param_to_list(&list, "-nbL",        "0", ARG_INTEGER,  &params->numBlev);
    add_param_to_list(&list, "-nb",         "0", ARG_INTEGER,  &params->num_b_frames);
    add_param_to_list(&list, "-me_rangeb",  "0", ARG_INTEGER,  &params->me_rangeb);
    add_param_to_list(&list, "-cIntra",     "0", ARG_INTEGER,  &params->constrainedIntraFlag);
    add_param_to_list(&list, "-multi_ref",  "0", ARG_INTEGER,  &params->multi_ref);
    add_param_to_list(&list, "-irPeriod",   "0", ARG_INTEGER,  &params->irPeriod);   
    add_param_to_list(&list, "-coarse_range_y",        "16", ARG_INTEGER,  &params->coarse_range_y);
    add_param_to_list(&list, "-coarse_range_x",        "128", ARG_INTEGER,  &params->coarse_range_x);
    add_param_to_list(&list, "-coarse_search_grid",        "1", ARG_INTEGER,  &params->coarse_search_grid);
    add_param_to_list(&list, "-coarse_search_use_dynamic",        "0", ARG_INTEGER,  &params->coarse_search_use_dynamic);
    add_param_to_list(&list, "-tr4flag",  "3", ARG_INTEGER,  &params->tr4flag);
    add_param_to_list(&list, "-tr16flag",  "3", ARG_INTEGER,  &params->tr16flag);    
    add_param_to_list(&list, "-lambda_coeff_SAD",  "1.10", ARG_FLOAT, &params->lambda_coeff_SAD);    
    add_param_to_list(&list, "-lambda_coeff_modeI",  "0.50", ARG_FLOAT, &params->lambda_coeff_modeI);
    add_param_to_list(&list, "-lambda_coeff_modeP",  "0.85", ARG_FLOAT, &params->lambda_coeff_modeP);
    add_param_to_list(&list, "-lambda_coeff_modeB",  "0.90", ARG_FLOAT, &params->lambda_coeff_modeB);
    add_param_to_list(&list, "-SADcost",  "0", ARG_INTEGER, &params->SADcost);
    add_param_to_list(&list, "-lambda_coeff_quantC",  "1.25", ARG_FLOAT, &params->lambda_coeff_quantC);
    add_param_to_list(&list, "-lambda_coeff_quantIntraI",  "0.50", ARG_FLOAT, &params->lambda_coeff_quantIntraI);
    add_param_to_list(&list, "-lambda_coeff_quantIntraPB",  "0.65", ARG_FLOAT,& params->lambda_coeff_quantIntraPB);
    add_param_to_list(&list, "-lambda_coeff_quantInterP",  "0.80", ARG_FLOAT, &params->lambda_coeff_quantInterP);
    add_param_to_list(&list, "-lambda_coeff_quantInterB",  "0.90", ARG_FLOAT, &params->lambda_coeff_quantInterB);
    add_param_to_list(&list, "-reset_contexts","0", ARG_INTEGER,  &params->resetContextsFlag);   
    add_param_to_list(&list, "-dqpP",      "4", ARG_INTEGER,  &params->dqpP);
    add_param_to_list(&list, "-pPeriod",   "1", ARG_INTEGER,  &params->pPeriod);
    add_param_to_list(&list, "-early_skip_thr",      "0", ARG_INTEGER,  &params->early_skip_thr);
	add_param_to_list(&list, "-bi",      "0", ARG_INTEGER,  &params->bi);
    add_param_to_list(&list, "-qp2",      "0", ARG_INTEGER,  &params->qp2);
    add_param_to_list(&list, "-qp2start", "-1", ARG_INTEGER,  &params->qp2start);


    /* Generate "argv" and "argc" for default parameters */
    default_argc = 1;
    for (i = 0; i < list.num; i++)
    {
        if (list.params[i].default_string != NULL)
        {
            default_argv[default_argc++] = list.params[i].name;
            default_argv[default_argc++] = list.params[i].default_string;
        }
    }

    /* Parse default parameters */
    parse_params(default_argc, default_argv, params, &list);

    /* Parse parameters from command line and config files */
    if (parse_params(argc, argv, params, &list) < 0)
        return NULL;

    if(params->num_frames <= 0)
    {
      fatalerror("Number of frames specified is 0 or negative");
    }

    if (params->me_rangeb == 0)
    {
        params->me_rangeb = params->me_range/2;
    }

    params->num_pixels = params->cols*params->rows;

    if (params->numBlev && (params->num_frames < (1<<params->numBlev))) {
        params->numBlev = 0;
        printf("Warning: Number of B levels forced to          0\n");
    }
    if (params->numBlev){
        if (params->num_b_frames != (1<<params->numBlev)-1)
            printf("Warning: Number of B frames forced to       %4d\n",
                    (1<<params->numBlev)-1);
        params->num_b_frames = (1<<params->numBlev)-1;
    }
    if (params->num_b_frames){

    }

    /* Increase the intra_period to the closest possible multiple of sub-gop if needed */
    if(params->intra_period % (params->num_b_frames+1) != 0)
    {
      params->intra_period += params->num_b_frames + 1 - (params->intra_period % (params->num_b_frames+1));
      printf("Warning: Intra period forced to %4d\n", params->intra_period);
    }


    if (params->pPeriod<1){
        params->pPeriod = 1;
        printf("Warning: P frame period must be positive, pPeriod set to 1\n");
    }
    if (params->pPeriod >1 && params->num_b_frames){
        params->pPeriod = 1;
        printf("Warning: Hierarchical P frames not supported with B frames, pPeriod set to 1\n");
    }
    
    if (params->me_type != ME_TYPE_DIAMOND && params->me_type != ME_TYPE_FULL_SEARCH)
		params->me_type = ME_TYPE_DIAMOND;

    return params;
}

/* Delete config params */
void delete_config_params(enc_params *params)
{
    if (params->infilestr != NULL)
        free(params->infilestr);

    if (params->outfilestr != NULL)
        free(params->outfilestr);

    if (params->reconfilestr != NULL)
        free(params->reconfilestr);

    if (params->statfilestr != NULL)
        free(params->statfilestr);

    free(params);
}
