/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include "../enc/global.h"
#include "rfutil.h"

#if 1
#define RF_END 0
#define FREE_ARG char*

static int num_memory_allocations = 0;

int get_num_allocations()
{
    return num_memory_allocations;
}

void *rf_malloc(size_t size)
{
    num_memory_allocations++;
    return malloc(size);
}

/*
void *rf_calloc(size_t num, size_t size)
{
    num_memory_allocations++;   
    return calloc(num, size);
#endif 
}
*/
void rf_free(void *memblock)
{
    num_memory_allocations--;
    free(memblock);    
}
#else

#define RF_END 1
#define FREE_ARG char*

static int num_memory_allocations = 0;

int get_num_allocations()
{
    return num_memory_allocations;
}

void *rf_malloc(size_t size)
{
    num_memory_allocations++;
    return malloc(size);
}
/*
void *rf_calloc(size_t num, size_t size)
{
    num_memory_allocations++;
    return calloc(num, size);
}
*/
void rf_free(void *memblock)
{
    num_memory_allocations--;
    free(memblock);
}
#endif

void rferror(char error_text[])
{
    fprintf(stderr,"Run-time error...\n");
    fprintf(stderr,"%s\n",error_text);
    fprintf(stderr,"...now exiting to system...\n");
    exit(1);
}

float *fvector(long nl, long nh)
     /* Allocate a float vector with subscript range v[nl..nh] */
{
    float *v;

    v=(float *)rf_malloc((size_t) ((nh-nl+1+RF_END)*sizeof(float)));
    if (!v) rferror("allocation failure in vector()");
    return v-nl+RF_END;
}

int *ivector(long nl, long nh)
     /* Allocate an int vector with subscript range v[nl..nh] */
{
    int *v;

    v=(int *)rf_malloc((size_t) ((nh-nl+1+RF_END)*sizeof(int)));
    if (!v) rferror("allocation failure in vector()");
    return v-nl+RF_END;
}

unsigned int *uivector(long nl, long nh)
     /* Allocate an unsigned int vector with subscript range v[nl..nh] */
{
    unsigned int *v;

    v=(unsigned int *)rf_malloc((size_t) ((nh-nl+1+RF_END)*sizeof(unsigned int)));
    if (!v) rferror("allocation failure in vector()");
    return v-nl+RF_END;
}

unsigned char *ucvector(long nl, long nh)
     /* Allocate a byte vector with subscript range v[nl..nh] */
{
    unsigned char *v;

    v = (unsigned char *)rf_malloc((size_t) ((nh-nl+1+RF_END)*
                                          sizeof(unsigned char)));
    if (!v) rferror("allocation failure in vector()");
    return v-nl+RF_END;
}

unsigned short *usvector(long nl, long nh)
     /* Allocate a short vector with subscript range v[nl..nh] */
{
    unsigned short *v;

    v = (unsigned short *)rf_malloc((size_t)((nh-nl+1+RF_END)*
                                          sizeof(unsigned short)));
    if (!v) rferror("allocation failure in vector()");
    return v-nl+RF_END;
}

short *svector(long nl, long nh)
     /* Allocate a short vector with subscript range v[nl..nh] */
{
    short *v;

    v = (short *)rf_malloc((size_t)((nh-nl+1+RF_END)*sizeof(short)));
    if (!v) rferror("allocation failure in vector()");
    return v-nl+RF_END;
}

unsigned char *ucsubvector(unsigned char *a,long oldnl,long oldnh,long newnl)
     /* point a subvecor [newnl..] to a[oldnl..oldnh] */
{
    long ncol=oldnl-newnl;
    unsigned char *m;

    /* set pointer  */
    m=a+ncol;

    /* return pointer to array of pointers to row */
    return m;
}

float **fmatrix(long nrl, long nrh, long ncl, long nch)
/* allocate a float matrix with subscript range m[nrl..nrh][ncl..nch] */
{
    long i,nrow=nrh-nrl+1,ncol=nch-ncl+1;
    float **m;

    /* allocate pointers to rows */
    m=(float **)rf_malloc((size_t)((nrow+RF_END)*sizeof(float*)));
    if (!m) rferror("allocation failure 1 in matrix()");
    m+=RF_END;
    m-=nrl;

    /* allocate rows and set pointers to them */
    m[nrl]=(float *)rf_malloc((size_t)((nrow*ncol+RF_END)*sizeof(float)));
    if (!m[nrl]) rferror("allocation failure 2 in matrix()");
    m[nrl]+=RF_END;
    m[nrl]-=ncl;

    for (i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

    /* return pointer to array of pointers to rows */
    return m;
}

double **dmatrix(long nrl, long nrh, long ncl, long nch)
/* allocate a double matrix with subscript range m[nrl..nrh][ncl..nch] */
{
    long i,nrow=nrh-nrl+1,ncol=nch-ncl+1;
    double **m;

    /* allocate pointers to rows */
    m=(double **)rf_malloc((size_t)((nrow+RF_END)*sizeof(double*)));
    if (!m) rferror("allocation failure 1 in matrix()");
    m+=RF_END;
    m-=nrl;

    /* allocate rows and set pointers to them */
    m[nrl]=(double *)rf_malloc((size_t)((nrow*ncol+RF_END)*sizeof(double)));
    if (!m[nrl]) rferror("allocation failure 2 in matrix()");
    m[nrl]+=RF_END;
    m[nrl]-=ncl;

    for (i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

    /* return pointer to array of pointers to rows */
    return m;
}

int **imatrix(long nrl, long nrh, long ncl, long nch)
     /* allocate an int matrix with subscript range m[nrl..nrh][ncl..nch] */
{
    long i,nrow=nrh-nrl+1,ncol=nch-ncl+1;
    int **m;

    /* allocate pointers to rows */
    m=(int **)rf_malloc((size_t)((nrow+RF_END)*sizeof(int*)));
    if (!m) rferror("allocation failure 1 in matrix()");
    m+=RF_END;
    m-=nrl;

    /* allocate rows and set pointers to them */
    m[nrl]=(int *)rf_malloc((size_t)((nrow*ncol+RF_END)*sizeof(int)));
    if (!m[nrl]) rferror("allocation failure 2 in matrix()");
    m[nrl]+=RF_END;
    m[nrl]-=ncl;

    for (i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

    /* return pointer to array of pointers to rows */
    return m;
}

unsigned char **ucmatrix(long nrl, long nrh, long ncl, long nch)
     /* allocate an unsigned char matrix with subscript range
        m[nrl..nrh][ncl..nch] */
{
    long i,nrow=nrh-nrl+1,ncol=nch-ncl+1;
    unsigned char **m;

    /* allocate pointers to rows */
    m=(unsigned char **)
        rf_malloc((size_t)((nrow+RF_END)*sizeof(unsigned char*)));
    if (!m) rferror("allocation failure 1 in matrix()");
    m+=RF_END;
    m-=nrl;

    /* allocate rows and set pointers to them */
    m[nrl]=(unsigned char *)
        rf_malloc((size_t)((nrow*ncol+RF_END)*sizeof(unsigned char)));
    if (!m[nrl]) rferror("allocation failure 2 in matrix()");
    m[nrl]+=RF_END;
    m[nrl]-=ncl;

    for (i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

    /* return pointer to array of pointers to rows */
    return m;
}    

unsigned short **usmatrix(long nrl, long nrh, long ncl, long nch)
     /* allocate an unsigned short matrix with subscript range
        m[nrl..nrh][ncl..nch] */
{
    long i,nrow=nrh-nrl+1,ncol=nch-ncl+1;
    unsigned short **m;

    /* allocate pointers to rows */
    m=(unsigned short **)
        rf_malloc((size_t)((nrow+RF_END)*sizeof(unsigned short*)));
    if (!m) rferror("allocation failure 1 in matrix()");
    m+=RF_END;
    m-=nrl;

    /* allocate rows and set pointers to them */
    m[nrl]=(unsigned short *)
        rf_malloc((size_t)((nrow*ncol+RF_END)*sizeof(unsigned short)));
    if (!m[nrl]) rferror("allocation failure 2 in matrix()");
    m[nrl]+=RF_END;
    m[nrl]-=ncl;

    for (i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

    /* return pointer to array of pointers to rows */
    return m;
}

short **smatrix(long nrl, long nrh, long ncl, long nch)
     /* allocate an unsigned short matrix with subscript range
        m[nrl..nrh][ncl..nch] */
{
    long i,nrow=nrh-nrl+1,ncol=nch-ncl+1;
    short **m;

    /* allocate pointers to rows */
    m=(short **)rf_malloc((size_t)((nrow+RF_END)*sizeof(short*)));
    if (!m) rferror("allocation failure 1 in matrix()");
    m+=RF_END;
    m-=nrl;

    /* allocate rows and set pointers to them */
    m[nrl]=(short *)rf_malloc((size_t)((nrow*ncol+RF_END)*sizeof(short)));
    if (!m[nrl]) rferror("allocation failure 2 in matrix()");
    m[nrl]+=RF_END;
    m[nrl]-=ncl;

    for (i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

    /* return pointer to array of pointers to rows */
    return m;
}

float **fsubmatrix(float **a, long oldrl, long oldrh, long oldcl,
                   long oldch, long newrl, long newcl)
     /* point a submatrix [newrl..][newcl..] to
        a[oldrl..oldrh][oldcl..oldch] */
{
    long i,j,nrow=oldrh-oldrl+1,ncol=oldcl-newcl;
    float **m;

    /* allocate array of pointers to rows */
    m=(float **)rf_malloc((size_t)((nrow+RF_END)*sizeof(float*)));
    if (!m) rferror("allocation failure in submatrix()");
    m+=RF_END;
    m-=newrl;

    /* set pointers to rows */
    for (i=oldrl,j=newrl;i<=oldrh;i++,j++) m[j]=a[i]+ncol;

    /* return pointer to array of pointers to row */
    return m;
}

int **isubmatrix(int **a, long oldrl, long oldrh, long oldcl,
                 long oldch, long newrl, long newcl)
     /* point a submatrix [newrl..][newcl..] to
        a[oldrl..oldrh][oldcl..oldch] */
{
    long i,j,nrow=oldrh-oldrl+1,ncol=oldcl-newcl;
    int **m;

    /* allocate array of pointers to rows */
    m=(int **)rf_malloc((size_t)((nrow+RF_END)*sizeof(int*)));
    if (!m) rferror("allocation failure in submatrix()");
    m+=RF_END;
    m-=newrl;

    /* set pointers to rows */
    for (i=oldrl,j=newrl;i<=oldrh;i++,j++) m[j]=a[i]+ncol;

    /* return pointer to array of pointers to row */
    return m;
}

void move_isubmatrix(int **m,int **a,
                     long oldrl,long oldrh,long oldcl,
                     long oldch, long newrl, long newcl)
     /* point a submatrix [newrl..][newcl..] to
        a[oldrl..oldrh][oldcl..oldch] */
{
    long i,j,ncol=oldcl-newcl;

    /* set pointers to rows */
    for (i=oldrl,j=newrl;i<=oldrh;i++,j++) m[j]=a[i]+ncol;
}

unsigned char **ucsubmatrix(unsigned char **a, long oldrl, long oldrh,
                            long oldcl, long oldch, long newrl, long newcl)
     /* point a submatrix [newrl..][newcl..] to
        a[oldrl..oldrh][oldcl..oldch] */
{
    long i,j,nrow=oldrh-oldrl+1,ncol=oldcl-newcl;
    unsigned char **m;

    /* allocate array of pointers to rows */
    m=(unsigned char **)rf_malloc((size_t)((nrow+RF_END)*sizeof(unsigned char*)));
    if (!m) rferror("allocation failure in submatrix()");
    m+=RF_END;
    m-=newrl;

    /* set pointers to rows */
    for (i=oldrl,j=newrl;i<=oldrh;i++,j++) m[j]=a[i]+ncol;

    /* return pointer to array of pointers to row */
    return m;
}

void move_ucsubvector(unsigned char *m,unsigned char *a,long oldnl,long oldnh,long newnl)
     /* point a subvector [newnl..] to a[oldnl..oldnh] */
{
    long ncol=oldnl-newnl;

    /* set pointer  */
    m=&a[0]+ncol;
}

void move_ucsubmatrix(unsigned char **m,unsigned char **a,
                     long oldrl,long oldrh,long oldcl,
                     long oldch, long newrl, long newcl)
     /* point a submatrix [newrl..][newcl..] to
        a[oldrl..oldrh][oldcl..oldch] */
{
    long i,j,ncol=oldcl-newcl;

    /* set pointers to rows */
    for (i=oldrl,j=newrl;i<=oldrh;i++,j++) m[j]=a[i]+ncol;
}

void free_fvector(float *v, long nl, long nh)
     /* free a float vector allocated with fvector() */
{
    rf_free((FREE_ARG) (v+nl-RF_END));
}

void free_ivector(int *v, long nl, long nh)
     /* free an int vector allocated with ivector() */
{
    rf_free((FREE_ARG) (v+nl-RF_END));
}

void free_uivector(unsigned int *v, long nl, long nh)
     /* free an unsigned int vector allocated with ivector() */
{
    rf_free((FREE_ARG) (v+nl-RF_END));
}

void free_ucvector(unsigned char *v, long nl, long nh)
     /* free an unsigned char vector allocated with ucvector() */
{
    rf_free((FREE_ARG) (v+nl-RF_END));
}

void free_usvector(unsigned short *v, long nl, long nh)
     /* free an unsigned short vector allocated with usvector() */
{
    rf_free((FREE_ARG) (v+nl-RF_END));
}

void free_svector(short *v, long nl, long nh)
     /* free a short vector allocated with svector() */
{
    rf_free((FREE_ARG) (v+nl-RF_END));
}

void free_fmatrix(float **m, long nrl, long nrh, long ncl, long nch)
     /* free a float matrix allocated by fmatrix() */
{
    rf_free((FREE_ARG) (m[nrl]+ncl-RF_END));
    rf_free((FREE_ARG) (m+nrl-RF_END));
}

void free_dmatrix(double **m, long nrl, long nrh, long ncl, long nch)
     /* free a double matrix allocated by dmatrix() */
{
    rf_free((FREE_ARG) (m[nrl]+ncl-RF_END));
    rf_free((FREE_ARG) (m+nrl-RF_END));
}

void free_imatrix(int **m, long nrl, long nrh, long ncl, long nch)
     /* free an int matrix allocated by imatrix() */
{
    rf_free((FREE_ARG) (m[nrl]+ncl-RF_END));
    rf_free((FREE_ARG) (m+nrl-RF_END));
}

void free_ucmatrix(unsigned char **m, long nrl, long nrh, long ncl, long nch)
     /* free an unsigned char matrix allocated by ucmatrix() */
{
    rf_free((FREE_ARG) (m[nrl]+ncl-RF_END));
    rf_free((FREE_ARG) (m+nrl-RF_END));
}

void free_usmatrix(unsigned short **m, long nrl, long nrh, long ncl, long nch)
     /* free an unsigned short matrix allocated by usmatrix() */
{
    rf_free((FREE_ARG) (m[nrl]+ncl-RF_END));
    rf_free((FREE_ARG) (m+nrl-RF_END));
}

void free_smatrix(short **m, long nrl, long nrh, long ncl, long nch)
     /* free a short matrix allocated by smatrix() */
{
    rf_free((FREE_ARG) (m[nrl]+ncl-RF_END));
    rf_free((FREE_ARG) (m+nrl-RF_END));
}

void free_fsubmatrix(float **b, long nrl, long nrh, long ncl, long nch)
     /* free a float submatrix allocated by submatrix() */
{
    rf_free((FREE_ARG) (b+nrl-RF_END));
}

void free_isubmatrix(int **b, long nrl, long nrh, long ncl, long nch)
     /* free an int submatrix allocated by submatrix() */
{
    rf_free((FREE_ARG) (b+nrl-RF_END));
}

void free_ucsubvector(unsigned char *a,long newnl,long newnh)
     /* nothing to free */
{
}

void free_ucsubmatrix(unsigned char **b, long nrl, long nrh, long ncl, long nch)
     /* free an unsigned char submatrix allocated by submatrix() */
{
    rf_free((FREE_ARG) (b+nrl-RF_END));
}

void free_f3tensor(float ***t, long nrl, long nrh, long ncl, long nch,
                   long ndl, long ndh)
     /* free a float 3tensor allocated by f3tensor() */
{
    rf_free((FREE_ARG) (t[nrl][ncl]+ndl-RF_END));
    rf_free((FREE_ARG) (t[nrl]+ncl-RF_END));
    rf_free((FREE_ARG) (t+nrl-RF_END));
}

void free_i3tensor(int ***t, long nrl, long nrh, long ncl, long nch,
                   long ndl, long ndh)
     /* free an int 3tensor allocated by i3tensor() */
{
    rf_free((FREE_ARG) (t[nrl][ncl]+ndl-RF_END));
    rf_free((FREE_ARG) (t[nrl]+ncl-RF_END));
    rf_free((FREE_ARG) (t+nrl-RF_END));
}

void free_c3tensor(unsigned char ***t, long nrl, long nrh, long ncl, long nch,
                   long ndl, long ndh)
     /* free an unsigned char 3tensor allocated by c3tensor() */
{
    rf_free((FREE_ARG) (t[nrl][ncl]+ndl-RF_END));
    rf_free((FREE_ARG) (t[nrl]+ncl-RF_END));
    rf_free((FREE_ARG) (t+nrl-RF_END));
}
/*
int get_mem1Ddouble(double **array1D, int raws)
{
	int j;

	if((*array1D = (double*)rf_calloc(raws, sizeof(double))) == NULL)
		printf("get_mem1Ddouble: array3D\n!");

	for(j = 0; j < raws; j++)
		(*array1D)[j] = 0.0;
	return sizeof(double)*raws;
}
*/

/*!
 ************************************************************************
 * \brief
 *    Allocate 2D memory array -> float array2D[rows][columns]
 *
 * \par Output:
 *    memory size in bytes
 ************************************************************************/
/*
int get_mem2Ddouble(double ***array2D, int rows, int columns)
{
	int i,j;

	if((*array2D = (double**)rf_calloc(rows, sizeof(double*))) == NULL)
		printf("get_mem2Ddouble: array2D");

	if(((*array2D)[0] = (double* )rf_calloc(columns*rows,sizeof(double ))) == NULL)
		printf("get_mem3Ddouble: array2D");

	for(i=1;i<rows;i++)
		(*array2D)[i] = (*array2D)[i-1] + columns ;

	for(i = 0; i < rows; i++)
		for(j = 0; j < columns; j++)
			(*array2D)[i][j] = 0.0;
	return sizeof(double)*rows*columns;
}
*/
/*!
 ************************************************************************
 * \brief
 *    free 1D memory array
 *    which was alocated with get_mem1DFloat()
 ************************************************************************
 */
void free_mem1Ddouble(double *array1D)
{
	if (array1D)
		rf_free(array1D);
	else
		printf("free_mem1Ddouble: array1D");
}

/*!
 ************************************************************************
 * \brief
 *    free 2D memory array
 *    which was alocated with get_mem2DFloat()
 ************************************************************************
 */
void free_mem2Ddouble(double **array2D)
{
	if (array2D)
	{
		if (array2D[0])
			rf_free(array2D[0]);
		else
			printf("free_mem2Ddouble: array2D");

		rf_free(array2D);
	}
	else
	{
		printf("free_mem2DDouble: trying to free unused memory");
		exit(-1);
	}
}
