/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/

#if !defined(_COST_H_)
#define _COST_H_

#include "tntmenc.h"

/*-----------------------------------------------------------------------------------------
Function:   SAD_16x16
Purpose:    Calculate SAD for a 16x16 block
Input:      *a           - Pointer to first block
            *b           - Pointer to second block
            stridea      - stride of first block
            strideb      - stride of second block
Return:     sad          - SAD value
Parameters: None 
-------------------------------------------------------------------------------------------*/

static inline unsigned int SAD_16x16(const unsigned char *a,
                                     const unsigned char *b,
                                     const int stridea,
                                     const int strideb)
{
    int i,j;
    unsigned int sad = 0;
    
    for (i=0;i<16;i++)
        for (j=0;j<16;j++)
            sad += abs(a[i*stridea+j] - b[i*strideb+j]);                  
    return(sad);
}

/*-----------------------------------------------------------------------------------------
Function:   SAD_16x8
Purpose:    Calculate SAD for a 16x8 block
Input:      *a           - Pointer to first block
            *b           - Pointer to second block
            stridea      - stride of first block
            strideb      - stride of second block
Return:     sad          - SAD value
Parameters: None 
-------------------------------------------------------------------------------------------*/

static inline unsigned int SAD_16x8(const unsigned char *a,
                                    const unsigned char *b,
                                    const int stridea,
                                    const int strideb)
{
    int i,j;
    unsigned int sad = 0;
    
    for (i=0;i<8;i++)
        for (j=0;j<16;j++)
            sad += abs(a[i*stridea+j] - b[i*strideb+j]);                  
    return(sad);
}

/*-----------------------------------------------------------------------------------------
Function:   SAD_8x16
Purpose:    Calculate SAD for a 8x16 block
Input:      *a           - Pointer to first block
            *b           - Pointer to second block
            stridea      - stride of first block
            strideb      - stride of second block
Return:     sad          - SAD value
Parameters: None 
-------------------------------------------------------------------------------------------*/

static inline unsigned int SAD_8x16(const unsigned char *a,
                                    const unsigned char *b,
                                    const int stridea,
                                    const int strideb)
{
    unsigned int sad = 0;
#if FAST_MODE
    int j;
    unsigned char a_[128], b_[128];
    
    for (j=0;j<16;j++)
        memcpy(&a_[8*j],&a[j*stridea],8);
    for (j=0;j<16;j++)
        memcpy(&b_[8*j],&b[j*strideb],8);
    for (j=0;j<128;j++)
        sad += abs(a_[j] - b_[j]);       
#else
    int i,j;
    for (i=0;i<16;i++)
        for (j=0;j<8;j++)
            sad += abs(a[i*stridea+j] - b[i*strideb+j]);   
#endif
    return(sad);
}

/*-----------------------------------------------------------------------------------------
Function:   SAD_8x8
Purpose:    Calculate SAD for a 8x8 block
Input:      *a           - Pointer to first block
            *b           - Pointer to second block
            stridea      - stride of first block
            strideb      - stride of second block
Return:     sad          - SAD value
Parameters: None 
-------------------------------------------------------------------------------------------*/

static inline unsigned int SAD_8x8(const unsigned char *a,
                                   const unsigned char *b,
                                   const int stridea,
                                   const int strideb)
{
    unsigned int sad = 0;
#if FAST_MODE
    int j;
    unsigned char a_[64], b_[64];
    
    for (j=0;j<8;j++)
        memcpy(&a_[8*j],&a[j*stridea],8);
    for (j=0;j<8;j++)
        memcpy(&b_[8*j],&b[j*strideb],8);
    for (j=0;j<64;j++)
        sad += abs(a_[j] - b_[j]);  
#else
    int i,j;
    for (i=0;i<8;i++)
        for (j=0;j<8;j++)
            sad += abs(a[i*stridea+j] - b[i*strideb+j]);   
#endif
    return(sad);
}

/*-----------------------------------------------------------------------------------------
Function:   SAD_32x32
Purpose:    Calculate SAD for a 32x32 block
Input:      *a           - Pointer to first block
            *b           - Pointer to second block
            stridea      - stride of first block
            strideb      - stride of second block
Return:     sad          - SAD value
Parameters: None 
-------------------------------------------------------------------------------------------*/
static inline unsigned int SAD_32x32(const unsigned char *a,
                                     const unsigned char *b,
                                     const int stridea,
                                     const int strideb)
{
    int i,j;
    unsigned int sad = 0;
#if FAST_MODE
    
    for (i=0;i<32;i++)
    {   
        unsigned char b_[32];   
        for (j=0;j<32;j++)
            b_[j] = b[i*strideb+j];

        for (j=0;j<32;j++)
            sad += abs(a[stridea*i+j] - b_[j]);
    }
#else
    for (i=0;i<32;i++)
        for (j=0;j<32;j++)
            sad += abs(a[stridea*i+j] - b[i*strideb+j]);
#endif
    return(sad);
}
/*-----------------------------------------------------------------------------------------
Function:   SAD_64x64
Purpose:    Calculate SAD for a 64x64 block
Input:      *a           - Pointer to first block
            *b           - Pointer to second block
            stridea      - stride of first block
            strideb      - stride of second block
Return:     sad          - SAD value
Parameters: None 
-------------------------------------------------------------------------------------------*/
static inline unsigned int SAD_64x64(const unsigned char *a,
                                     const unsigned char *b,
                                     const int stridea,
                                     const int strideb)
{
    int i,j;
    unsigned int sad = 0;
#if FAST_MODE
    
    for (i=0;i<64;i++)
    {   
        unsigned char b_[64];   
        for (j=0;j<64;j++)
            b_[j] = b[i*strideb+j];

        for (j=0;j<64;j++)
            sad += abs(a[stridea*i+j] - b_[j]);
    }
#else
    for (i=0;i<64;i++)
        for (j=0;j<64;j++)
            sad += abs(a[stridea*i+j] - b[i*strideb+j]);
#endif
    return(sad);
}
/*-----------------------------------------------------------------------------------------
Function:   SAD_NxN
Purpose:    Calculate SAD for a NxN block
Input:      *a           - Pointer to first block
            *b           - Pointer to second block
            stridea      - stride of first block
            strideb      - stride of second block
            mode         - macrolock mode
Return:     sad          - SAD value
Parameters: None 
-------------------------------------------------------------------------------------------*/

static inline unsigned int SAD_NxN(const unsigned char *a,
                                   const unsigned char *b,
                                   const int stridea,
                                   const int strideb, 
                                   const mb_mode mode)
{
    unsigned int sad;
    switch (mode)
    {
      case MODE_16x16 : sad = SAD_16x16(a,b,stridea,strideb); 
          break;
      case MODE_16x8:   sad = SAD_16x8(a,b,stridea,strideb); 
          break;
      case MODE_8x16:   sad = SAD_8x16(a,b,stridea,strideb); 
          break;
      case MODE_8x8:    sad = SAD_8x8(a,b,stridea,strideb); 
          break;
      case MODE_LMB32:  sad = SAD_32x32(a,b,stridea,strideb); 
          break;
      case MODE_LMB64:  sad = SAD_64x64(a,b,stridea,strideb); 
          break;
    }
    return(sad);
}
/*-----------------------------------------------------------------------------------------
Function:   SAD_NxN_LMB
Purpose:    Calculate SAD for a 32x32 or 64x64 block
Input:      *a           - Pointer to first block
            *b           - Pointer to second block
            stridea      - stride of first block
            strideb      - stride of second block
            mode         - macrolock mode
Return:     sad          - SAD value
Parameters: None 
-------------------------------------------------------------------------------------------*/

static inline unsigned int SAD_NxN_LMB(const unsigned char *a,
                                   const unsigned char *b,
                                   const int stridea,
                                   const int strideb, 
                                   const mb_mode mode)
{
    unsigned int sad;
    switch (mode)
    {
      case MODE_LMB32:  sad = SAD_32x32(a,b,stridea,strideb); 
          break;
      case MODE_LMB64:  sad = SAD_64x64(a,b,stridea,strideb); 
          break;
    }
    return(sad);
}

/*-----------------------------------------------------------------------------------------
Function:   SSD_NxN
Purpose:    Calculate SSD for an MxN block
Input:      *a           - Pointer to first block
            *b           - Pointer to second block
            width        - width of blocks
            height       - height of blocks
Return:     sad          - SAD value
Parameters: None 
-------------------------------------------------------------------------------------------*/

static inline int SSD_NxN(const unsigned char *a,const unsigned char *b,const int width,const int height)
{    
	int i,ssdval = 0;
    for (i=0;i<height*width;i++)
    {
        ssdval += ((a[i]-b[i])*(a[i]-b[i])); 
    }
    return ssdval;
}

/*-----------------------------------------------------------------------------------------
Function:   cost_MB
Purpose:    Calculate SSD-baed or SAD-based RD cost for a 16x16 MB
Input:      bits         - Number of bits used
            lambda       - Lagrange multiplier
            L1_or_L2     - SAD (1) or SSD (2)
Return:     cost         - SSD/SAD + lambda*bits
Parameters: None 
-------------------------------------------------------------------------------------------*/

static inline int cost_MB(macroblock_info *macroblockInfo,
                          const int bits,
                          const double lambda,
                          const int L1_or_L2)
{
    int cost;
    switch (L1_or_L2)
    {
    case 1:
        cost  = SAD_16x16(macroblockInfo->orgY,macroblockInfo->recY,16,16);
        cost += SAD_8x8(macroblockInfo->orgU,macroblockInfo->recU,8,8);
        cost += SAD_8x8(macroblockInfo->orgV,macroblockInfo->recV,8,8);  
        break;
    case 2:
        cost  = SSD_NxN(macroblockInfo->orgY,macroblockInfo->recY,16,16);
        cost += SSD_NxN(macroblockInfo->orgU,macroblockInfo->recU,8,8);
        cost += SSD_NxN(macroblockInfo->orgV,macroblockInfo->recV,8,8);
        break;
    }
    cost += (int)(lambda*bits + 0.5f);    
    return cost;
}

/*-----------------------------------------------------------------------------------------
Function:   cost_LMB
Purpose:    Calculate SSD-baed or SAD-based RD cost for a large MB
Input:      width        - Width of block
            height       - Height of block
            bits         - Number of bits used
            lambda       - Lagrange multiplier
            L1_or_L2     - SAD (1) or SSD (2)
            mode         - MB mode
Return:     cost         - SSD/SAD + lambda*bits
Parameters: None 
-------------------------------------------------------------------------------------------*/

static inline int cost_LMB(macroblock_info *macroblockInfo,
                           const int width,
                           const int height,
                           const int bits,
                           const double lambda,
                           const int L1_or_L2,
                           const mb_mode mode)
{

    int cost;

    switch (L1_or_L2)
    {
    case 1:
        switch (mode)
        {
        case MODE_LMB32 :
            cost = SAD_32x32(macroblockInfo->orgLMBY,macroblockInfo->recLMBY,32,32);
            cost += SAD_16x16(macroblockInfo->orgLMBU,macroblockInfo->recLMBU,16,16);
            cost += SAD_16x16(macroblockInfo->orgLMBV,macroblockInfo->recLMBV,16,16);       
            break;
        case MODE_LMB64 :
            cost = SAD_64x64(macroblockInfo->orgLMBY,macroblockInfo->recLMBY,64,64);
            cost += SAD_32x32(macroblockInfo->orgLMBU,macroblockInfo->recLMBU,32,32);
            cost += SAD_32x32(macroblockInfo->orgLMBV,macroblockInfo->recLMBV,32,32);       
            break;
        }
        break;
    case 2:
       switch (mode)
        {
        case MODE_LMB32 :
            cost = SSD_NxN(macroblockInfo->orgLMBY,macroblockInfo->recLMBY,32,32);
            cost += SSD_NxN(macroblockInfo->orgLMBU,macroblockInfo->recLMBU,16,16);
            cost += SSD_NxN(macroblockInfo->orgLMBV,macroblockInfo->recLMBV,16,16);       
            break;
        case MODE_LMB64 :
            cost = SSD_NxN(macroblockInfo->orgLMBY,macroblockInfo->recLMBY,64,64);
            cost += SSD_NxN(macroblockInfo->orgLMBU,macroblockInfo->recLMBU,32,32);
            cost += SSD_NxN(macroblockInfo->orgLMBV,macroblockInfo->recLMBV,32,32);       
            break;
        }
        break;
    }
    cost += (int)(lambda*bits + 0.5f);    
    return cost;
}

/*-----------------------------------------------------------------------------------------
Function:   cost_block
Purpose:    Calculate SSD-baed or SAD-based RD cost for a block
Input:      *a           - Pointer to first block
            *b           - Pointer to second block
            width        - Width of block
            height       - Height of block
            bits         - Number of bits used
            lambda       - Lagrange multiplier
            L1_or_L2     - SAD (1) or SSD (2)
            mode         - MB mode
Return:     cost         - SSD/SAD + lambda*bits
Parameters: None 
-------------------------------------------------------------------------------------------*/

static inline int cost_block(unsigned char *a, unsigned char *b,int width,int height,int bits,double lambda,int L1_or_L2,mb_mode mode)
{
    int cost;
    switch (L1_or_L2)
    {
    case 1:
        cost = SAD_NxN(a,b,width,height,mode);
        break;
    case 2:
        cost = SSD_NxN(a,b,width,height);     
        break;
    }
    cost += (int)(lambda*bits + 0.5f);    
    return cost;
}

/*-----------------------------------------------------------------------------------------
Function:   rd_cost_lookup
Purpose:    Calculate cost used for ME
Input:      qp        - quantizer value
            x         - Horizontal motion vector difference
            y         - Vertical motion vector difference
Return:               - lambda * MVbits
Parameters: None 
-------------------------------------------------------------------------------------------*/

static inline int rd_cost_lookup(const image_frame *f, int qp,int x,int y)
{
    return (int)(f->rd_lambda[qp]*(float)(f->mv_bits_center[x]+f->mv_bits_center[y])+0.5f);
}

#endif  // _COST_H_
