/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include "global.h"
#include "putbits.h"
//STATIC #include "rfutil.h"

/* No. of bytes to expand bitstream with when necessary*/
/* #define BCHUNK 2048       */
#define BCHUNK 2000000


static unsigned int mask[33] = {
    0x00000000,0x00000001,0x00000003,0x00000007,
    0x0000000f,0x0000001f,0x0000003f,0x0000007f,
    0x000000ff,0x000001ff,0x000003ff,0x000007ff,
    0x00000fff,0x00001fff,0x00003fff,0x00007fff,
    0x0000ffff,0x0001ffff,0x0003ffff,0x0007ffff,
    0x000fffff,0x001fffff,0x003fffff,0x007fffff,
    0x00ffffff,0x01ffffff,0x03ffffff,0x07ffffff,
    0x0fffffff,0x1fffffff,0x3fffffff,0x7fffffff,
    0xffffffff};

/*-----------------------------------------------------------------------------------------
Function:   initbits
Purpose:    Initialize bitstream buffer (before first call to putbits or alignbits)
Input:      outfile    - The output stream contains a pointer to the outputfile
Output:     The output stream
Parameters: None 
-------------------------------------------------------------------------------------------*/

stream *initbits(FILE *outfile)
{
    stream *str;

    if ((str = (stream *)malloc((size_t)sizeof(stream))) == NULL)
    {
        fatalerror("Problem allocating bitstream buffer.");
    }

    if ((str->bitstream = (unsigned char *)malloc((size_t)(BCHUNK*sizeof(unsigned char)))) == NULL)
    {
        fatalerror("Problem allocating bitstream buffer.");
    }
    str->bytesize = BCHUNK;
    str->bytepos = 0;
    str->bytecount = 0;
    str->bitbuf = 0;
    str->bitrest = 32;
    str->outfile = outfile;
    return str;
}

/*-----------------------------------------------------------------------------------------
Function:   flush_all_bits
Purpose:    Writing all bits to file, including padding 
            Should only be called at end of completed encoding
Input:      str    - The output stream 
Output:     None
Parameters: None 
-------------------------------------------------------------------------------------------*/
void flush_all_bits(stream *str)
{
    int i;
    int bytes = 4 - str->bitrest/8;

    if ((str->bytepos+bytes) > str->bytesize)
    {
        flush_bytebuf(str);
    }
    for (i = 0; i < bytes; i++)
    {
        str->bitstream[str->bytepos++] = (str->bitbuf >> (24-i*8)) & 0xff;
    }

    if (str->outfile)
    {
        if (fwrite(str->bitstream,sizeof(unsigned char),str->bytepos,str->outfile) != str->bytepos)
        {
            fatalerror("Problem writing bitstream to file.");
        }
    }
    str->bytepos = 0;
}    
                    
/*-----------------------------------------------------------------------------------------
Function:   flush_bytebuf
Purpose:    Writing bytes from bitstream to file
Input:      str    - The output stream 
Output:     None
Parameters: None 
-------------------------------------------------------------------------------------------*/
void flush_bytebuf(stream *str)
{
    if (str->outfile)
    {
        if (fwrite(str->bitstream,sizeof(unsigned char),str->bytepos,str->outfile) != str->bytepos)
        {
            fatalerror("Problem writing bitstream to file.");
        }
    }
    str->bytepos = 0;
}

/*-----------------------------------------------------------------------------------------
Function:   flush_bitbuf
Purpose:    Writing bits from the bitbuffer to the bitstream .
            Called as soon as the bitbuffer (32 bits) is full.
Input:      str    - The output stream 
Output:     None
Parameters: None 
-------------------------------------------------------------------------------------------*/
void flush_bitbuf(stream *str)
{
    if ((str->bytepos+4) > str->bytesize)
    {
        fatalerror("Run out of bits in stream buffer.");
    }
    str->bitstream[str->bytepos++] = (str->bitbuf >> 24) & 0xff;
    str->bitstream[str->bytepos++] = (str->bitbuf >> 16) & 0xff;
    str->bitstream[str->bytepos++] = (str->bitbuf >> 8) & 0xff;
    str->bitstream[str->bytepos++] = str->bitbuf & 0xff;
    str->bytecount += 4;
    str->bitbuf = 0;
    str->bitrest = 32;
}

/*-----------------------------------------------------------------------------------------
Function:   putbits
Purpose:    Putting bits into the oputput stream.
Input:      n      - Number of bits to put
            val    - Binary value of length n bits, that will be put into the stream 
            str    - The output stream 
Output:     None
Parameters: None 
-------------------------------------------------------------------------------------------*/
void putbits(unsigned int n, unsigned int val, stream *str)
{
    unsigned int rest;

    if (n <= str->bitrest)
    {
        str->bitbuf |= ((val & mask[n]) << (str->bitrest-n));
        str->bitrest -= n;
    }
    else
    {
        rest = n-str->bitrest;
        str->bitbuf |= (val >> rest) & mask[n-rest];
        flush_bitbuf(str);
        str->bitbuf |= (val & mask[rest]) << (32-rest);
        str->bitrest -= rest;
    }
}

/*-----------------------------------------------------------------------------------------
Function:   bitcount
Purpose:    Give the total number of bits put into the output stream
Input:      str    - The output stream 
Output:     Total number of bits put into the output stream
Parameters: None 
-------------------------------------------------------------------------------------------*/
int bitcount(stream *str)
{
  return 8*str->bytecount + (32-str->bitrest);
}


/*-----------------------------------------------------------------------------------------
Function:   closebits
Purpose:    Free the memory allocated by the output stream
Input:      str    - The output stream 
Output:     None
Parameters: None 
-------------------------------------------------------------------------------------------*/
void closebits(stream *str)
{
    free(str->bitstream);
    free(str);
}


/*-----------------------------------------------------------------------------------------
Function:   copyCodedStream
Purpose:    Make a copy of an output stream
Input:      srcStream    - The output stream that should by copied
Output:     destStream   - A copy of the srcStream
Parameters: None 
-------------------------------------------------------------------------------------------*/
void copyCodedStream(stream * destStream, stream *srcStream)
{
	memcpy(destStream->bitstream,srcStream->bitstream, BCHUNK*sizeof(unsigned char));
    destStream->bytesize = srcStream->bytesize;
    destStream->bytepos = srcStream->bytepos;
    destStream->bytecount = srcStream->bytecount;
    destStream->bitbuf = srcStream->bitbuf;
    destStream->bitrest = srcStream->bitrest;

}

